def get_enumerator_type(enumerator):
    '''
    Define the type of enumerator
    
    An enumerator may be hashtag, arabic, alpha, or roman 
    upper or lower
    period, full_parenthesis, or right_parenthesis
    
    For example: "II)" is an  ["roman", "upper", "right_parenthesis"]
    '''
    
    output = []
    
    if enumerator.endswith('.'):
        output.append('period')
        enumerator = enumerator[0:-1]
    elif enumerator.startswith('('):
        output.append('full_parenthesis')
        enumerator = enumerator[1:-1]
    else:
        output.append('right_parenthesis')
        enumerator = enumerator[0:-1]
    
    try:
        int(enumerator)
        output.append('arabic')
    
    except ValueError:
        if enumerator == '#':
            output.append('hashtag')
        
        elif enumerator.lower() == 'i':
            output.append('roman')
            if enumerator.isupper():
                output.append('upper')
            else:
                output.append('lower')
        
        elif len(enumerator) == 1:
            output.append('alpha')
            if enumerator.isupper():
                output.append('upper')
            else:
                output.append('lower')
        
        else:
            output.append('roman')
            if enumerator.isupper():
                output.append('upper')
            else:
                output.append('lower')
    return output



import string
from .roman import toRoman

def make_enumerator(enumerator_type, enumerator_level):
    '''Make an enumerator given the type and level'''
    enumerator = ''
    
    if 'hashtag' in enumerator_type:
        enumerator = '#'
    
    elif 'arabic' in enumerator_type:
        enumerator = str(enumerator_level)
    
    elif 'alpha' in enumerator_type:
        enumerator = string.ascii_lowercase[enumerator_level - 1]
    
    elif 'roman' in enumerator_type:
        enumerator = toRoman(enumerator_level)
    
    if 'upper' in enumerator_type:
        enumerator = enumerator.upper()
    
    elif 'lower' in enumerator_type:
        enumerator = enumerator.lower()
    
    if 'period' in enumerator_type:
        enumerator = enumerator + '.'
    
    elif 'full_parenthesis' in enumerator_type:
        enumerator = '(' + enumerator + ')'
    
    elif 'right_parenthesis' in enumerator_type:
        enumerator = enumerator + ')'
    
    return enumerator






def collect_paragraph(lines, index):
    output = [lines[index]]
    leading_space = lines[index].replace(lines[index].lstrip(), '')
    index += 1
    
    while index < len(lines):
        if lines[index].startswith(leading_space) and not lines[index] == '':
            output.append(lines[index].strip())
            index += 1
        else:
            return ' '.join(output), index
    return ' '.join(output), index



def is_definition_term(lines, index):
    '''Verify if the line is a definition term'''
    if index == len(lines) - 1:
        return False
        
    leading_space = lines[index].replace(lines[index].lstrip(), '')
    if lines[index + 1].startswith(leading_space + '    '):
        return True
    



def is_directive(lines, index):
    '''check if the line contains a directive'''
    if lines[index].startswith('.. ') and '::' in lines[index]:
        return True
    
    #Handle hyperlinks here
    if lines[index].startswith('.. _'):
        return True
    
    return False



from .is_option_list_item import is_option_list_item
from .tools import space_fill

def collect_option_list_item(lines, index, interspace):
    output = []
    leading_space = lines[index].replace(lines[index].lstrip(), '')
    
    option = lines[index].lstrip().split('  ')[0]
    lspace = leading_space + space_fill(len(option), ' ') + '  '
    
    rest_of_text = lines[index].lstrip()[len(option)::].lstrip()
    
    output.append(leading_space + option + interspace + rest_of_text)
    index += 1
    
    while index < len(lines):
        if not is_option_list_item(lines, index):
            if not lines[index] == '':
                output.append(lines[index].lstrip())
                index += 1
            elif index < len(lines) - 1 and lines[index + 1].startswith(lspace):
                output.append('\n\n')
                output.append(lspace + lines[index])
                index += 1
            else:
                return ' '.join(output), index
        else:
            return ' '.join(output), index
    
    return ' '.join(output), index



from .tools.is_only import is_only

def is_title(lines, index):
    '''check to see if the line is a title'''
    if index > len(lines) - 2:
        return False
    if len(lines[index + 1]) == 0:
        return False
    symbol = lines[index + 1][0]
    
    if is_only(lines[index + 1], [symbol]):
        if len(lines[index]) <= len(lines[index + 1]):
            return True
    
    return False



from .reflow_paragraph import reflow_paragraph
from .tools import space_fill

def reflow_option_list_item(text, space):
    
    leading_space = text.replace(text.lstrip(), '')
    option = text.lstrip().split('  ')[0]
    rest_of_text = text.lstrip()[len(option)::]
    interspace = rest_of_text.replace(rest_of_text.lstrip(), '')
    
    lspace = leading_space + space_fill(len(option), ' ') + interspace
    
    blocks = rest_of_text.split('\n')
    for b in range(len(blocks)):
        blocks[b] = reflow_paragraph(blocks[b].lstrip(), space, lspace).lstrip()
    
    
    for b in range(len(blocks)):
        if b == 0:
            blocks[b] = leading_space + option + interspace + blocks[b]
        else:
            blocks[b] = lspace + blocks[b]
    
    return '\n'.join(blocks)



from .roman import fromRoman
import string

def get_enumerator_level(enumerator, enumerator_type):
    enumerator = enumerator[0:-1]
    
    if 'hashtag' in enumerator_type:
        return 0
    
    if 'full_parenthesis' in enumerator_type:
        enumerator = enumerator[1::]
    
    if 'arabic' in enumerator_type:
        return int(enumerator)
    
    elif 'alpha' in enumerator_type:
        return string.ascii_lowercase.index(enumerator.lower()) + 1
    
    else:
        return fromRoman(enumerator)



import math

def getSpan(spans, row, column):
    """checks if a row,column is in spans"""
    for i in range(len(spans)):
        if [row, column] in spans[i]:
            return spans[i]
    return None


def lineBreak(count, symbol):
    """makes a string that is count long of symbol"""
    x = ""
    for i in range(0, count):
        x = x + symbol
    return x

def getSpanColumnCount(span):
    """Gets the number of columns inluded in a span"""
    columns = 1
    first_column = span[0][1]
    for i in range(len(span)):
        if span[i][1] > first_column:
            columns += 1
            first_column = span[i][1]
    return columns

def sortSpans(spans):
    """Ensure the first cell of each span is the text cell"""
    for span in range(len(spans)):
        spans[span] = sorted(spans[span])
    return spans

def addCushions(table):
    """adds space to start and end of each item in a list of lists"""
    for row in range(len(table)):
        for column in range(len(table[row])):
            lines = table[row][column].split("\n")
            for i in range(len(lines)):
                if not lines[i] == "":
                    lines[i] = " " + lines[i].rstrip() + " "
            table[row][column] = "\n".join(lines)
    return table

def isOnly(text, char_list):
    text = text.strip()
    if text == '':
        return False
    for i in range(len(text)):
        if text[i] not in char_list:
            return False
    return True

class TableTypeError(Exception):
    """Raised if a Table is not a List of Lists"""

class TableContentError(Exception):
    """Raised if a Table is empty or has inconsistent column counts"""

class SpanError(Exception):
    """
    Raised if the spans in any spangroup are not touching or 
    the span is not square
    """
    

class Cell():
    """Holds the text and data for an rst text cell"""
    def __init__(self, text, row, column, row_count, column_count):
        self.text = text
        self.row = row
        self.column = column
        self.row_count = row_count
        self.column_count = column_count
    
    @property
    def left_sections(self):
        lines = self.text.split('\n')
        sections = 0
        for i in range(len(lines)):
            if lines[i].startswith('+'):
                sections += 1
        return sections - 1
    
    @property
    def right_sections(self):
        lines = self.text.split('\n')
        sections = 0
        for i in range(len(lines)):
            if lines[i].endswith('+'):
                sections += 1 
        return sections - 1
    
    @property
    def top_sections(self):
        top_line = self.text.split('\n')[0]
        return len(top_line.split('+')) - 2
    
    @property
    def bottom_sections(self):
        bottom_line = self.text.split('\n')[-1]
        return len(bottom_line.split('+')) - 2
    
    @property
    def is_header(self):
        bottom_line = self.text.split('\n')[-1]
        return isOnly(bottom_line, ['+', '='])

    def h_center_cell(self):
        lines = self.text.split('\n')
        cell_width = len(lines[0]) - 2
        
        truncated_lines = []
        for i in range(1, len(lines) - 1):
            truncated = lines[i][2:len(lines[i]) - 2].rstrip()
            truncated_lines.append(truncated)

        truncated_lines.append('')
        truncated_lines.insert(0, '')
        
        max_line_length = len(max(truncated_lines, key=len))
        remainder = cell_width - max_line_length
        left_width = math.floor(remainder / 2)
        left_space = lineBreak(left_width, ' ')

        for i in range(len(truncated_lines)):
            truncated_lines[i] = left_space + truncated_lines[i]
            right_width = cell_width - len(truncated_lines[i])
            truncated_lines[i] += lineBreak(right_width, ' ')
            
        
        for i in range(1, len(lines) - 1):
            lines[i] = lines[i][0] + truncated_lines[i] + lines[i][-1]
        
        self.text = '\n'.join(lines)
        self.v_center_cell()
        
    def v_center_cell(self):
        lines = self.text.split('\n')
        cell_width = len(lines[0]) - 2
        
        truncated_lines = []
        for i in range(1, len(lines) - 1):
            truncated = lines[i][1:len(lines[i]) - 1]
            truncated_lines.append(truncated)
        
        total_height = len(truncated_lines)
        
        above_trash_count = 0
        for i in range(len(truncated_lines)):
            if truncated_lines[i].rstrip() == '':
                above_trash_count += 1
            else:
                break
        
        below_trash_count = 0
        for i in reversed(range(len(truncated_lines))):
            if truncated_lines[i].rstrip() == '':
                below_trash_count += 1
            else:
                break
                
        significant_lines = truncated_lines[above_trash_count: len(truncated_lines) - below_trash_count]
        
        remainder = total_height - len(significant_lines)
        blank = lineBreak(cell_width, ' ')
        above_height = math.floor(remainder / 2)
        for i in range(0, above_height):
            significant_lines.insert(0, blank)
        
        below_height = math.ceil(remainder / 2)
        for i in range(0, below_height):
            significant_lines.append(blank)
        
        significant_lines.insert(0, '')
        significant_lines.append('')
        
        for i in range(1, len(lines) - 1):
            lines[i] = lines[i][0] + significant_lines[i] + lines[i][-1]
        
        self.text = '\n'.join(lines)
            

    def __lt__(self, other):
        """For sorting instances of this class"""
        return [self.row, self.column] < [other.row, other.column]

    def mergeableDirection(self, other):
        """Determines the direction in which two cells can be merged"""
        self_left = self.column
        self_right = self.column + self.column_count
        self_top = self.row
        self_bottom = self.row + self.row_count

        other_left = other.column
        other_right = other.column + other.column_count
        other_top = other.row
        other_bottom = other.row + other.row_count

        if (self_right == other_left and self_top == other_top and
                self_bottom == other_bottom and self.right_sections >= other.left_sections):
            return "RIGHT"
        elif (self_left == other_left and self_right == other_right and
                self_top == other_bottom and self.top_sections >= other.bottom_sections):
            return "TOP"
        elif (self_left == other_left and
              self_right == other_right and
              self_bottom == other_top and self.bottom_sections >= other.top_sections):
            return "BOTTOM"
        elif (self_left == other_right and
              self_top == other_top and
              self_bottom == other_bottom and self.left_sections >= other.right_sections):
            return "LEFT"
        else:
            return "NONE"

    def merge(self, other):
        """attempts to merge two cells"""

        self_lines = self.text.split("\n")
        other_lines = other.text.split("\n")

        if self.mergeableDirection(other) == "RIGHT":
            for i in range(len(self_lines)):
                self_lines[i] = self_lines[i] + other_lines[i][1::]
            self.text = "\n".join(self_lines)
            self.column_count += other.column_count
            return True
        elif self.mergeableDirection(other) == "TOP":
            if self_lines[0].count('+') > other_lines[-1].count('+'):
                other_lines.pop(-1)
            else:
                self_lines.pop(0)
            other_lines.extend(self_lines)
            self.text = "\n".join(other_lines)
            self.row_count += other.row_count
            self.row = other.row
            self.column = other.column
            return True
        elif self.mergeableDirection(other) == "BOTTOM":
            if self_lines[-1].count('+') > other_lines[0].count('+') or self.is_header:
                other_lines.pop(0)
            else:
                self_lines.pop(-1)
            self_lines.extend(other_lines)
            self.text = "\n".join(self_lines)
            self.row_count += other.row_count
            return True
        elif self.mergeableDirection(other) == "LEFT":
            for i in range(len(self_lines)):
                self_lines[i] = other_lines[i][0:-1] + self_lines[i]
            self.text = "\n".join(self_lines)
            self.column_count += other.column_count
            self.row = other.row
            self.column = other.column
            return True
        else:
            return False

def getLongestLineLength(text):
    """Get the length longest line in a paragraph"""
    lines = text.split("\n")
    length = 0
    for i in range(len(lines)):
        if len(lines[i]) > length:
            length = len(lines[i])
    return length


def getSpanRowCount(span):
    """Gets the number of rows included in a span"""
    rows = 1
    first_row = span[0][0]
    for i in range(len(span)):
        if span[i][0] > first_row:
            rows += 1
            first_row = span[i][0]
    return rows

def getSpanColumnCount(span):
    """Gets the number of rows included in a span"""
    first_column = span[0][1]
    column = first_column
    for pair in span:
        if pair[1] > column:
            column = pair[1]
    return (column - first_column) + 1
        
def getTotalSpanHeight(span, heights):
    """Sum the row heights of a span"""
    start_row = span[0][0]
    row_count = getSpanRowCount(span)
    total_height = 0
    for i in range(start_row, start_row + row_count):
        total_height += heights[i]
    total_height += row_count - 1
    return total_height


def getTotalSpanWidth(span, widths):
    """Sum the widths of a span"""
    start_column = span[0][1]
    column_count = getSpanColumnCount(span)
    total_width = 0
    for i in range(start_column, start_column + column_count):
        total_width += widths[i]
    total_width += column_count - 1
    return total_width


def mergeCells(cells):
    """Loop through list of cells and piece them together one by one"""
    current = 0
    while len(cells) > 1:
        count = 0
        while count < len(cells):
            if cells[current].merge(cells[count]):
                if current > count:
                    current -= 1
                cells.pop(count)
            else:
                count += 1
        current += 1
        if current >= len(cells):
            current = 0
    return cells[0].text


def makeTextCell(table, span, widths, heights, use_headers):
    """Creates an rst text Cell"""
    width = getTotalSpanWidth(span, widths)
    height = getTotalSpanHeight(span, heights)
    text_row = span[0][0]
    text_column = span[0][1]
    text = table[text_row][text_column]

    lines = text.split("\n")
    for i in range(len(lines)):
        width_difference = width - len(lines[i])
        lines[i] = lines[i] + lineBreak(width_difference, " ")

    height_difference = height - len(lines)
    empty_lines = []
    for i in range(0, height_difference):
        empty_lines.append(lineBreak(width, " "))
    lines.extend(empty_lines)

    output = ["+" + lineBreak(width, "-") + "+"]
    for i in range(0, height):
        output.append("|" + lines[i] + "|")

    if use_headers and span[0][0] == 0:
        symbol = "="
    else:
        symbol = "-"
    output.append("+" + lineBreak(width, symbol) + "+")

    text = "\n".join(output)
    row_count = getSpanRowCount(span)
    column_count = getSpanColumnCount(span)
    cell = Cell(text, text_row, text_column, row_count, column_count)

    return cell


def getHeights(table, spans):
    """get the heights of the rows of the output table"""
    span_remainders = {}
    for span in spans:
        span_remainders[str(span)] = 0

    heights = []
    for row in table:
        heights.append(-1)
    
    for row in range(len(table)):
        for column in range(len(table[row])):
            text = table[row][column]
            span = getSpan(spans, row, column)
            row_count = getSpanRowCount(span)
            height = len(text.split('\n'))
            if row_count == 1 and height > heights[row]:
                heights[row] = height

    for row in range(len(table)):
        for column in range(len(table[row])):
            span = getSpan(spans, row, column)
            row_count = getSpanRowCount(span)
            if row_count > 1:
                text_row = span[0][0]
                text_column = span[0][1]
                text = table[text_row][text_column]
                height = len(text.split('\n')) - (row_count - 1)
                add_row = 0
                while height > sum(heights[text_row:text_row + row_count]):
                    heights[text_row + add_row] += 1
                    if add_row + 1 < row_count:
                        add_row += 1
                    else:
                        add_row = 0
    return heights

def getWidths(table, spans):
    """Get the widths of the columns of the output table"""

    widths = []
    for column in table[0]:
        widths.append(3)
    
    for row in range(len(table)):
        for column in range(len(table[row])):
            span = getSpan(spans, row, column)
            column_count = getSpanColumnCount(span)
            if column_count == 1:
                text_row = span[0][0]
                text_column = span[0][1]
                text = table[text_row][text_column]
                length = getLongestLineLength(text)
                if length > widths[column]:
                    widths[column] = length

    for row in range(len(table)):
        for column in range(len(table[row])):
            span = getSpan(spans, row, column)
            column_count = getSpanColumnCount(span)
            if column_count > 1:
                text_row = span[0][0]
                text_column = span[0][1]
                text = table[text_row][text_column]
                length = getLongestLineLength(text)
                end_column = text_column + column_count
                available_space = sum(widths[text_column: end_column])
                available_space += column_count - 1
                while length > available_space:
                    for i in range(text_column, end_column):
                        widths[i] += 1
                        available_space = sum(widths[text_column: end_column])
                        available_space += column_count - 1
                        if length <= available_space:
                            break
    return widths


def convertToSpans(table, spans):
    """Converts all cells to spans"""
    new_spans = []
    for row in range(len(table)):
        for column in range(len(table[row])):
            span = getSpan(spans, row, column)
            if not span:
                new_spans.append([[row, column]])
    new_spans.extend(spans)
    new_spans = list(sorted(new_spans))
    return new_spans

def checkTable(table):
    """Checks to make sure that each row has the same number of columns"""
    if not type(table) is list:
        raise(TableTypeError("Table must be a list of lists"))
    if len(table) == 0:
        raise(TableContentError("Table must contain at least one row and one column"))
    for i in range(len(table)):
        if not type(table[i]) is list:
            raise(TableTypeError("Table must be a list of lists of strings"))
        if not len(table[i]) == len(table[0]):
            raise(TableContentError("Each row must have the same number of columns"))

    
def checkSpan(span):
    """Check the spans in a spangroup"""
    if not type(span) is list:
        raise(SpanError("Spans must be a list/tuple of lists"))
    for pair in span:
        if not type(pair) is list:
            raise(SpanError("Spans must be a list/tuple of lists"))
        if not len(pair) == 2:
            raise(SpanError("Spans must be a [Row, Column] pair of integers"))
    
    total_rows = getSpanRowCount(span)
    total_columns = getSpanColumnCount(span)
    
    if not len(span) == total_rows * total_columns:
        raise SpanError('Somethings wrong with the span group: ' + str(span))
    
    checked = [span.pop(0)]

    while len(span) > 0:
        row = span[0][0]
        col = span[0][1]
        matched = False
        
        for i in range(len(checked)):
            if row == checked[i][0] and abs(col - checked[i][1]) == 1:
                matched = True
            
            elif abs(row - checked[i][0]) == 1 and col == checked[i][1]:
                matched = True
            
        if matched:
            checked.append(span.pop(0))
        
        else:
            checked.extend(span)
            raise SpanError('Somethings wrong with the span group: ' + str(checked))

def data2rst(table, spans=[[[0, 0]]], use_headers=True, center_cells=False, center_headers=False):
    checkTable(table)
    
    for row in range(len(table)):
        for column in range(len(table[row])):
            table[row][column] = str(table[row][column])

    table = addCushions(table)
    spans = sortSpans(spans)
    
    for span in spans:
        span = list(span)
        checkSpan(span)
    
    spans = convertToSpans(table, spans)

    widths = getWidths(table, spans)
    heights = getHeights(table, spans)

    cells = []
    for span in spans:
        cell = makeTextCell(table, span, widths, heights, use_headers)
        cells.append(cell)
    cells = list(sorted(cells))
    
    if center_cells:
        for cell in cells:
            if not cell.is_header:
                cell.h_center_cell()
    
    if center_headers:
        for cell in cells:
            if cell.is_header:
                cell.h_center_cell()
            
    output = mergeCells(cells)

    return output



from .is_field import is_field

def collect_field(lines, index, interspace):
    output = []
    words = lines[index].strip().split(' ')
    for x in range(len(words)):
        if words[x].endswith(':') and not words[x].endswith('\\:'):
            field_name = ' '.join(words[0:x + 1])
            words = words[x + 1::]
            break
    
    leading_space = lines[index].replace(lines[index].lstrip(), '')
    first_line = leading_space + field_name + interspace + ' '.join(words).lstrip()
    output.append(first_line.rstrip())
    
    index += 1

    while index < len(lines):
        if (not is_field(lines, index) and
            not lines[index] == ''
        ):
            output.append(lines[index].lstrip())
            index += 1
        else:
            return ' '.join(output), index
    
    return ' '.join(output), index



def is_bullet_list_item(lines, index):
    '''check if it's a bullet list item'''
    
    bullets = ['*', '-', '+', '•', '‣', '⁃']
    first_char = lines[index].lstrip().split(' ')[0]
    if first_char in bullets:
        return True
    return False
        



def collect_gridtable(lines, index):
    output = []
    while (index < len(lines) and 
            (lines[index].lstrip().startswith('+') or 
             lines[index].lstrip().startswith('|'))
          ):
        output.append(lines[index])
        index += 1
    return '\n'.join(output), index



def is_only(text, char_list):
    '''
    return true if the text is made up of only the characters in char 
    list
    '''
    text = text.strip()
    if text == '':
        return False
    for i in range(len(text)):
        if text[i] not in char_list:
            return False
    return True



def reflow_paragraph(text, space, leading_space=''):
    '''
    Reflow a flattened paragraph so it fits inside horizontal 
    space
    '''
    words = text.split(' ')
    growing_string = leading_space
    output_list = []
    
    while len(words) > 0:
        if growing_string == leading_space:
            growing_string += words[0]
            words.pop(0)
        elif len(growing_string + ' ' + words[0]) <= space:
            growing_string += ' ' + words[0]
            words.pop(0)
        else:
            output_list.append(growing_string + '\n')
            growing_string = leading_space
    output_list.append(growing_string)
    return ''.join(output_list)



from .is_bullet_list_item import is_bullet_list_item
from .is_enumerated_list_item import is_enumerated_list_item
from .is_todo_list_item import is_todo_list_item

from .tools import make_enumerator
from .tools import space_fill

def collect_enumerated_list_item(lines, index, max_ewidth):
    '''Collect an enumerated list item'''
    output= []
    leading_space = lines[index].replace(lines[index].lstrip(), '')
    enumerator = lines[index].lstrip().split(' ')[0]
    initial = len(leading_space) + len(enumerator)
    rest_of_text = lines[index][initial::].lstrip()
    
    espace = space_fill(len(enumerator), ' ')
    interspace = space_fill(max_ewidth - len(enumerator), ' ')
    
    lines[index] = leading_space + enumerator + interspace + ' ' + rest_of_text
    output.append(lines[index])
    index += 1
    
    lspace = leading_space + espace + interspace
    while index < len(lines):
        if (lines[index].startswith(lspace) and
          not is_bullet_list_item(lines, index) and
          not is_enumerated_list_item(lines, index) and
          not is_todo_list_item(lines, index)
        ):
            output.append(lines[index].lstrip())
            index += 1
        else:
            return ' '.join(output), index
    
    return ' '.join(output), index



from .tools import space_fill

def reflow_enumerated_list_item(text, space):
    '''
    reflow an enumerated list item
    '''
    leading_space = text.replace(text.lstrip(), '')
    lines = text.split('\n')
    enumerator = text.lstrip().split(' ')[0]
    initial = len(leading_space) + len(enumerator)
    rest_of_text = text[initial::]
    interspace = rest_of_text.replace(rest_of_text.lstrip(), '')
    words = [enumerator]
    words.extend(rest_of_text.lstrip().split(' '))
    growing_string = leading_space
    
    output_list = []
    while len(words) > 0:
        if len(growing_string) == len(leading_space):
            growing_string += words[0]
            words.pop(0)
        elif len(growing_string) == len(leading_space) + + len(enumerator):
            growing_string += interspace + words[0]
            words.pop(0)
        elif len(growing_string + ' ' + words[0]) <= space:
            growing_string += ' ' + words[0]
            words.pop(0)
        else:
            output_list.append(growing_string + '\n')
            growing_string = leading_space + space_fill(len(enumerator), ' ')
    output_list.append(growing_string + '\n')
    return ''.join(output_list).rstrip()

    return '\n'.join(lines)



def other_colon_present(words):
    for word in words:
        if word.endswith(':') and not word.endswith('\\:'):
            return True
    return False

def is_field(lines, index):
    words = lines[index].lstrip().split(' ')

    if not words[0].endswith(':'):
        if not words[0].startswith(':'):
            return False
        
        if not other_colon_present(words):
            return False
            
        else:
            return True
    
    leading_space = lines[index].replace(lines[index].lstrip(), '')
    
    if (index < len(lines) - 1 and 
        not is_field(lines, index + 1) and 
        not lines[index + 1].startswith(leading_space + ' ') and 
        not lines[index + 1] == ''):
        return False
    
    return True



from .is_bullet_list_item import is_bullet_list_item
from .is_enumerated_list_item import is_enumerated_list_item
from .is_todo_list_item import is_todo_list_item

def collect_todo_list_item(lines, index):
    '''Collect a todo list item'''
    output = [lines[index]]
    leading_space = lines[index].replace(lines[index].lstrip(), '')
    index += 1
    
    while index < len(lines):
        if (
          lines[index].startswith(leading_space + ' ') and
          not lines[index] == '' and
          not is_bullet_list_item(lines, index) and
          not is_enumerated_list_item(lines, index) and
          not is_todo_list_item(lines, index)
        ):
            output.append(lines[index].lstrip())
            index += 1
        else:
            return ' '.join(output), index
    
    return ' '.join(output), index



from .is_footnote import is_footnote

def collect_footnote(lines, index):
    output = []
    output.append(lines[index])
    index += 1

    while index < len(lines):
        if (
          not lines[index] == '' and
          not is_footnote(lines, index)
        ):
            output.append(lines[index].lstrip())
            index += 1
        else:
            return ' '.join(output), index
    return ' '.join(output), index

lines = '.. [#Strange] this is a strangely wrapped footnote that will be rewrapped.'



from .reflow_paragraph import reflow_paragraph
#from .tools import space_fill

'''
def reflow_footnote(text, space):
    output = []
    leading_space = text.replace(text.lstrip(), '')
    ref = text.lstrip().split(']')[0]
    rest_of_text = text.lstrip()[len(ref)::]
    
    lspace = leading_space + space_fill(len(ref), ' ') + '  '
    paragraph = reflow_paragraph(rest_of_text, space, lspace)
    
    return leading_space + ref + paragraph.lstrip()
'''
def reflow_footnote(text, space):
    leading_space = text.replace(text.lstrip(), '')
    text = text[2::].lstrip()
    
    paragraph = reflow_paragraph(text, space, leading_space + '   ')
    return leading_space + '.. ' + paragraph.lstrip()



from .tools import get_enumerator_level
from .tools import get_enumerator_type
from .tools import make_enumerator
from .tools import space_fill

# These 2 modules come from DashTable
from .tools import grid2data
from .tools import data2rst

from .is_main_title import is_main_title
from .is_title import is_title
from .is_bullet_list_item import is_bullet_list_item
from .is_enumerated_list_item import is_enumerated_list_item
from .is_todo_list_item import is_todo_list_item
from .is_definition_term import is_definition_term
from .is_field import is_field
from .is_option_list_item import is_option_list_item
from .is_footnote import is_footnote
from .is_directive import is_directive
from .is_gridtable import is_gridtable

from .collect_main_title import collect_main_title
from .collect_title import collect_title
from .collect_bullet_list_item import collect_bullet_list_item
from .collect_enumerated_list_item import collect_enumerated_list_item
from .collect_todo_list_item import collect_todo_list_item
from .collect_field import collect_field
from .collect_option_list_item import collect_option_list_item
from .collect_footnote import collect_footnote
from .collect_directive import collect_directive
from .collect_gridtable import collect_gridtable
from .collect_paragraph import collect_paragraph

from .reflow_bullet_list_item import reflow_bullet_list_item
from .reflow_enumerated_list_item import reflow_enumerated_list_item
from .reflow_todo_list_item import reflow_todo_list_item
from .reflow_field import reflow_field
from .reflow_option_list_item import reflow_option_list_item
from .reflow_footnote import reflow_footnote
from .reflow_paragraph import reflow_paragraph

import itertools
import copy

def get_last_enumerator(lines, i):
    this_leading_space = lines[i].replace(lines[i].lstrip(), '')
    this_enumerator = lines[i].lstrip().split(' ')[0]
    this_enumerator_type = get_enumerator_type(this_enumerator)
    
    if i == 0:
        return this_enumerator
    elif i > 1 and lines[i - 1] == '' and lines[i - 2].replace(lines[i - 2].lstrip(), '') == this_leading_space:
        return this_enumerator
    
    for x in reversed(range(0, i)):
        if lines[x] == '' or lines[x].startswith(this_leading_space):
            if is_enumerated_list_item(lines, x):
                leading_space = lines[x].replace(lines[x].lstrip(), '')
                if leading_space == this_leading_space:
                    enumerator = lines[x].lstrip().split(' ')[0]
                    enumerator_type = get_enumerator_type(enumerator)
                    if enumerator_type == this_enumerator_type:
                        return enumerator
            elif not lines[x] == '' and not lines[x].startswith(this_leading_space + '  '):
                return this_enumerator
        else:
            return this_enumerator
    return this_enumerator

def renumerate_enumerators(lines):
    for i in range(len(lines)):
        if is_enumerated_list_item(lines, i):
            leading_space = lines[i].replace(lines[i].lstrip(), '')
            last_enumerator = get_last_enumerator(lines, i)
            
            enumerator = lines[i].lstrip().split(' ')[0]
            initial = len(leading_space) + len(enumerator)
            rest_of_text = lines[i][initial::]
            
            if not last_enumerator == enumerator:
                enumerator_type = get_enumerator_type(last_enumerator)
                enumerator_level = get_enumerator_level(last_enumerator, enumerator_type) + 1
            else:
                enumerator_type = get_enumerator_type(last_enumerator)
                enumerator_level = get_enumerator_level(last_enumerator, enumerator_type)
            
            new_enumerator = make_enumerator(enumerator_type, enumerator_level)
            lines[i] = leading_space + new_enumerator + rest_of_text
    return lines

def reflow(text, space=0):
    '''Reflow the text to fit within "space" number of characters'''
    lines = text.split('\n')
    for i in range(len(lines)):
        lines[i] = lines[i].rstrip()
    lines = renumerate_enumerators(lines)
        
    reflowed = []
    
    index = 0
    while index < len(lines):
        if lines[index] == '':
            reflowed.append('')
            index += 1
        elif is_main_title(lines, index):
            flat, index = collect_main_title(lines, index)
            reflowed.append(flat)
        elif is_title(lines, index):
            flat, index = collect_title(lines, index)
            reflowed.append(flat)
        elif is_gridtable(lines, index):
            text, index = collect_gridtable(lines, index)
            leading_space = text.replace(text.lstrip(), '')
            table, spans, use_headers = grid2data(text)
            
            cell_widths = get_cell_widths(table, spans, space - len(leading_space))
            for row in range(len(table)):
                for column in range(len(table[row])):
                    table[row][column] = reflow(table[row][column], cell_widths[row][column]).rstrip()
            gridtable = data2rst(table, spans, use_headers, center_headers=True)
            grid_lines = gridtable.split('\n')
            for i in range(len(grid_lines)):
                grid_lines[i] = leading_space + grid_lines[i]
            reflowed.append('\n'.join(grid_lines))
            
        elif is_bullet_list_item(lines, index):
            flat, index = collect_bullet_list_item(lines, index)
            if space:
                reflowed.append(reflow_bullet_list_item(flat, space))
            else:
                reflowed.append(flat)
        elif is_enumerated_list_item(lines, index):
            max_ewidth = 0
            last_leading_space = lines[i].replace(lines[i].lstrip(), '')
            for x in range(index, len(lines)):
                if lines[x] == '' or lines[x].startswith(last_leading_space):
                    if is_enumerated_list_item(lines, x):
                        leading_space = lines[x].replace(lines[x].lstrip(), '')
                        if leading_space == last_leading_space:
                            enumerator = lines[x].lstrip().split(' ')[0]
                            if len(enumerator) > max_ewidth:
                                max_ewidth = len(enumerator)
                    elif not lines[x] == '' and not lines[x].startswith(last_leading_space + '  '):
                        break
                else:
                    break
                    
            for x in reversed(range(0, len(reflowed))):
                if reflowed[x] == '' or reflowed[x].startswith(last_leading_space):
                    if is_enumerated_list_item(reflowed, x):
                        leading_space = reflowed[x].replace(reflowed[x].lstrip(), '')
                        if leading_space == last_leading_space:
                            enumerator = reflowed[x].lstrip().split(' ')[0]
                            if len(enumerator) > max_ewidth:
                                max_ewidth = len(enumerator)
                    elif not reflowed[x] == '' and not reflowed[x].startswith(last_leading_space + '  '):
                        break
                else:
                    break
            
            #print(max_ewidth)
            
            flat, index = collect_enumerated_list_item(
                lines, index, max_ewidth)
            if space:
                reflowed.append(reflow_enumerated_list_item(flat, space))
            else:
                reflowed.append(flat)
        elif is_todo_list_item(lines, index):
            flat, index = collect_todo_list_item(lines, index)
            if space:
                reflowed.append(reflow_todo_list_item(flat, space))
            else:
                reflowed.append(flat)
        
        elif is_field(lines, index):
            field_names = []
            leading_space = lines[index].replace(lines[index].lstrip(), '')
            for i in reversed(range(0, len(reflowed))):
                if is_field(reflowed, i):
                    words = reflowed[i].strip().split(' ')
                    for x in range(len(words)):
                        if words[x].endswith(':') and not words[x].endswith('\\:'):
                            field_name = ' '.join(words[0:x + 1])
                            field_names.append(field_name)
                else:
                    break
            
            current_field_name = ''
            for i in range(index, len(lines)):
                if is_field(lines, i):
                    words = lines[i].strip().split(' ')
                    for x in range(len(words)):
                        if words[x].endswith(':') and not words[x].endswith('\\:'):
                            field_name = ' '.join(words[0:x + 1])
                            field_names.append(field_name)
                            if i == index:
                                current_field_name = field_name
                elif not lines[i].startswith(leading_space + '    ') and not lines[i] == '':
                    break
            max_field_name_length = len(max(field_names, key=len))
            
            remainder = max_field_name_length - len(current_field_name)
            interspace = space_fill(remainder, ' ') + ' '
            
            flat, index = collect_field(lines, index, interspace)
            if space:
                reflowed.append(reflow_field(flat, space))
            else:
                reflowed.append(flat)
        
        elif is_option_list_item(lines, index):
            options = []
            leading_space = lines[index].replace(lines[index].lstrip(), '')
            for i in reversed(range(0, len(reflowed))):
                if is_option_list_item(reflowed, i):
                    options.append(reflowed[i].lstrip().split('  ')[0])
                else:
                    break
                    
            current_option = lines[index].lstrip().split('  ')[0]
            for i in range(index, len(lines)):
                if is_option_list_item(lines, i):
                    options.append(lines[i].lstrip().split('  ')[0])
                elif not lines[i].startswith(leading_space + ' ') and not lines[i] == '':
                    break
                elif lines[i] == '':
                    if i < len(lines) - 1 and not lines[i + 1].startswith(leading_space + ' '):
                        break
            max_option_width = len(max(options, key=len))
            remainder = max_option_width - len(current_option)
            interspace = space_fill(remainder, ' ') + '  '
            flat, index = collect_option_list_item(lines, index, interspace)
            if space:
                reflowed.append(reflow_option_list_item(flat, space))
            else:
                reflowed.append(flat)
        
        elif is_footnote(lines, index):
            flat, index = collect_footnote(lines, index)
            if space:
                reflowed.append(reflow_footnote(flat, space))
            else:
                reflowed.append(flat)
        
        elif is_directive(lines, index):
            flat, index = collect_directive(lines, index)
            reflowed.append(flat)
        
        elif is_definition_term(lines, index):
            flat = lines[index]
            reflowed.append(flat)
            index += 1
        
        else:
            leading_space = lines[index].replace(lines[index].lstrip(), '')
            flat, index = collect_paragraph(lines, index)
            if space:
                reflowed.append(reflow_paragraph(flat, space, leading_space))
            else:
                reflowed.append(flat)
    
    return '\n'.join(reflowed).rstrip()

def getSpanColumnCount(span):
    """Gets the number of columns inluded in a span"""
    columns = 1
    first_column = span[0][1]
    for i in range(len(span)):
        if span[i][1] > first_column:
            columns += 1
            first_column = span[i][1]
    return columns

def getSpanRowCount(span):
    """Gets the number of rows included in a span"""
    rows = 1
    first_row = span[0][0]
    for i in range(len(span)):
        if span[i][0] > first_row:
            rows += 1
            first_row = span[i][0]
    return rows
    
def grid_height_from_cols(table, spans, combo):
    '''Calculate the sum of the cell text heights given column widths'''
    row_heights = []
    for row in table:
        row_heights.append(1)
    
    for row in range(len(table)):
        for column in range(len(table[row])):
            text = table[row][column]
            space = combo[column]
            for span in spans:
                if [row, column] in span:
                    span_row_count = getSpanRowCount(span)
                    if not span_row_count > 1:
                        span_col_count = getSpanColumnCount(span)
                        span_row_count = getSpanRowCount(span)
                        r = span[0][0]
                        c = span[0][1]
                        if span_col_count > 1 and span_row_count == 1 and row == r:
                            text = table[r][c]
                            space = sum(combo[c: c + span_col_count + 1]) + (span_col_count - 1)
                    break
            reflowed = reflow(text, space)
            height = len(reflowed.split('\n'))
            if row_heights[row] < height:
                row_heights[row] = height
    
    for row in range(len(table)):
        for column in range(len(table[row])):
            for span in spans:
                if [row, column] in span:
                    span_row_count = getSpanRowCount(span)
                    if span_row_count > 1:
                        span_col_count = getSpanColumnCount(span)
                        r = span[0][0]
                        c = span[0][1]
                        if span_col_count > 1 and row == r:
                            text = table[r][c]
                            space = sum(combo[c: c + span_col_count + 1]) + (span_col_count - 1)
                    reflowed = reflow(text, space)
                    height = len(reflowed.split('\n'))
                    height = height - (span_col_count - 1)
                    add_row = 0
                    while height > sum(row_heights[r:r + span_row_count]):
                        row_heights[r + add_row] += 1
                        if add_row + 1 < len(row_heights):
                            add_row += 1
                        else:
                            add_row = 0
    return sum(row_heights)
    

def get_column_combo(height, col_height_width_dicts):
    '''Get the column widths as they would be if set at height'''
    combo = []
    for col in col_height_width_dicts:
        keys = list(sorted(col.keys()))
        for key in keys:
            if height >= key:
                addition = col[key]
        if type(addition).__name__ == 'int':
            combo.append(addition)
        elif type(addition).__name__ == 'tuple':
            combo.extend(addition)
    return combo

def reevaluate_max_col_widths(
    max_col_widths, min_col_widths, space):
    '''
    A column's max width must not be more than 
    allowed space - sum(all other min_col_widths)
    '''
    column_count = len(max_col_widths)
    text_space = space - (column_count + 1) - (2 * column_count)
    for i in range(len(max_col_widths)):
        min_sum = sum(min_col_widths[0:i]) + sum(min_col_widths[i + 1::])
        if max_col_widths[i] + min_sum > text_space:
            
            if text_space - min_sum >= min_col_widths[i]:
                max_col_widths[i] = text_space - min_sum
            else:
                max_col_widths[i] = min_col_widths[i]
    return max_col_widths

def get_breakpoints(text, min_width, max_width):
    '''
    Find where newlines could be added to a string
    but don't bother if the result would be longer than max width
    or shorter than min width
    '''
    breakpoints = []
    text = text.replace('\n', ' ')
    words = text.split(' ')
    growing_string = ''
    for word in words:
        growing_string += word + ' '
        if len(growing_string.rstrip()) >= min_width and len(growing_string.rstrip()) <= max_width:
            breakpoints.append(len(growing_string.rstrip()))
    return breakpoints

def make_cell_width_dict(
    table, row, column, spans, max_col_widths, min_col_widths):
    '''Create a dictionary of cell heights at specific widths'''
    cell_width_dict = {}
    text = table[row][column]
    
    for span in spans:
        if [row, column] in span:
            r = span[0][0]
            c = span[0][1]
            text = table[r][c]
    
    breakpoints = get_breakpoints(
        table[row][column], min_col_widths[column], max_col_widths[column])
    
    for bp in breakpoints:
        reflowed = reflow(text, bp)
        rlines = reflowed.split('\n')
        cell_width_dict[bp] = len(rlines)
    
    if len(breakpoints) == 0:
        cell_width_dict[min_col_widths[column]] = 1
    
    return cell_width_dict

def get_key_column_widths(
    table, spans, max_col_widths, min_col_widths):
    '''
    Discover which widths are significant to a column's reflowed height
    '''
    cell_width_dicts = []
    for row in range(len(table)):
        cell_width_dicts.append([])
        for column in range(len(table[row])):
            cell_width_dicts[-1].append(
                make_cell_width_dict(
                    table, row, column, spans, max_col_widths, min_col_widths))
    
    key_column_widths = []
    for column in max_col_widths:
        key_column_widths.append([])
    
    for row in range(len(cell_width_dicts)):
        for column in range(len(cell_width_dicts[row])):
            key_column_widths[column].extend(list(cell_width_dicts[row][column].keys()))
    
    for col in range(len(key_column_widths)):
        key_column_widths[col] = list(set(key_column_widths[col]))
    
    return cell_width_dicts, key_column_widths

def has_column_span(table, column, spans):
    '''
    Determine if a column has any rows containing a column-spanned cell
    If so, return the row containing the furthest longest column span
    '''
    rows = {}
    for row in range(len(table)):
        for span in spans:
            if [row, column] in span:
                span_col_count = getSpanColumnCount(span)
                if span_col_count > 1:
                    rows[span_col_count] = row
    if rows == {}:
        return None
    else:
        keys = list(rows.keys())
        return rows[max(keys)]
    
    return None

def make_column_height_width_dict(column, cell_width_dicts):
    '''
    For a non-column-spanned column, determine it's height at breakpoint
    widths and make a dictionary of the heights:widths
    '''
    breakpoints = []
    for row in range(len(cell_width_dicts)):
        for col in range(len(cell_width_dicts[row])):
            if col == column:
                breakpoints.extend(list(cell_width_dicts[row][col].keys()))
    breakpoints = list(set(breakpoints))
    height_widths = {}
    for bp in breakpoints:
        column_height = 0
        for row in range(len(cell_width_dicts)):
            for col in range(len(cell_width_dicts[row])):
                if column == col:
                    keys = list(sorted(cell_width_dicts[row][column].keys(), reverse=True))
                    added = False
                    for i in range(len(keys)):
                        if bp >= keys[i] and added == False:
                            column_height += cell_width_dicts[row][column][keys[i]]
                            added = True
                    if added == False:
                        column_height += cell_width_dicts[row][column][keys[-1]] + 1
        try:
            if bp < height_widths[column_height]:
                height_widths[column_height] = bp
        except KeyError:
            height_widths[column_height] = bp
    return height_widths
    
def get_last_column(table, column, furthest, spans):
    '''
    This is useful for when a grid has multiple column spans that
    overlap eachother. We need the group of columns that share
    overlapping column spans.
    '''
    for row in range(len(table)):
        for col in range(len(table[row])):
            if col >= column and col <= furthest:
                long_row = has_column_span(table, column, spans)
                if long_row and long_row > furthest:
                    furthest = get_last_column(table, col, long_row, spans)
    return furthest
    
def in_spans(row, column, spans):
    for span in spans:
        if [row, column] in span:
            return True
    return False
    
def make_spanned_column_height_width_dict(
    table, min_column_widths, column, long_row, spans, max_column_widths, 
    cell_width_dicts, text_space):
    '''
    For a group of columns united by column spans, go through all
    possible combinations of widths to arrive at the few that would give
    the best possible height for those widths.
    
    If you have a grid with lots of column spans, this function is the
    reason why the script is running slow.
    '''
    for span in spans:
        if [long_row, column] in span:
            span_col_count = getSpanColumnCount(span)
            furthest = span[0][1] + span_col_count - 1
    last_column = get_last_column(table, column, furthest, spans)
    
    included_columns = list(range(column, last_column + 1))
    
    key_column_widths = []
    for column in max_column_widths:
        key_column_widths.append([])
    
    for row in range(len(cell_width_dicts)):
        for col in range(len(cell_width_dicts[row])):
            if col in included_columns:
                if not in_spans(row, col, spans):
                    key_column_widths[col].extend(list(cell_width_dicts[row][col].keys()))
    
    col = 0
    while col < len(key_column_widths):
        if key_column_widths[col] == []:
            key_column_widths.pop(col)
        else:
            key_column_widths[col] = list(set(key_column_widths[col]))
            col += 1
    
    combos = list(itertools.product(*key_column_widths))
        
    i = 0
    while i < len(combos):
        if sum(combos[i]) >= text_space:
            combos.pop(i)
        else:
            i += 1
    
    adjusted_table = []
    adjusted_spans = []
    for row in range(len(table)):
        adjusted_table.append([])
        for col in range(len(table[row])):
            if col in included_columns:
                adjusted_table[-1].append(table[row][col])
            for span in spans:
                if [row, col] in span:
                    if not span in adjusted_spans:
                        new_span = copy.deepcopy(span)
                        adjusted_spans.append(new_span)
    
    for i in range(len(adjusted_spans)):
        for p in range(len(adjusted_spans[i])):
            adjusted_spans[i][p][1] -= 1
    
    adjusted_min_column_widths = []
    for col in range(len(min_column_widths)):
        if col in included_columns:
            adjusted_min_column_widths.append(min_column_widths[col])
    
    smallest_height = grid_height_from_cols(adjusted_table, adjusted_spans, adjusted_min_column_widths)
    height_width_dict = {}
    
    for i in range(len(combos)):
        height = grid_height_from_cols(adjusted_table, adjusted_spans, combos[i])
        try:
            if sum(combos[i]) < sum(height_width_dict[height]):
                height_width_dict[height] = combos[i]
        except KeyError:
            height_width_dict[height] = combos[i]
    
    return height_width_dict, column + span_col_count

def make_final_cell_widths(table, spans, column_widths):
    '''
    Determine the final width of each cell. This should be the same
    as the column widths for normal cells, but will be different for
    column-spanded cells
    '''
    cell_widths = []
    for row in range(len(table)):
        cell_widths.append([])
        for column in range(len(table[row])):
            cell_widths[-1].append(column_widths[column])
    
    for row in range(len(table)):
        for column in range(len(table[row])):
            width = column_widths[column]
            for span in spans:
                if [row, column] in span:
                    span_col_count = getSpanColumnCount(span)
                    if span_col_count > 1:
                        r = span[0][0]
                        c = span[0][1]
                        width = sum(column_widths[c: c + span_col_count + 1]) + (span_col_count - 1)
            cell_widths[row][column] = width
    return cell_widths

def get_combo_widths(combo, col_height_width_dicts):
    cols = []
    for c in range(len(combo)):
        if type(col_height_width_dicts[c][combo[c]]).__name__ == 'int':
            cols.append(col_height_width_dicts[c][combo[c]])
        else:
            cols.extend(col_height_width_dicts[c][combo[c]])
    return cols

def get_cell_widths(table, spans, space):
    '''Get the new cell widths for the reflowed grid table'''
    
    min_cell_widths = []
    max_cell_widths = []
    for row in range(len(table)):
        min_cell_widths.append([])
        max_cell_widths.append([])
        for column in range(len(table[row])):
            span_col_count = 1
            for span in spans:
                if [row, column] in span:
                    span_col_count = getSpanColumnCount(span)
            if span_col_count == 1:
                min_text = reflow(table[row][column], 1)
                max_text = reflow(table[row][column], 0)

                min_lines = min_text.split('\n')
                min_cell_widths[row].append(len(max(min_lines, key=len)))
                
                max_lines = max_text.split('\n')
                max_cell_widths[row].append(len(max(max_lines, key=len)))
            else:
                min_cell_widths[row].append(0)
                max_cell_widths[row].append(space)
    
    for row in range(len(table)):
        for column in range(len(table[row])):
            for span in spans:
                if [row, column] in span:
                    span_col_count = getSpanColumnCount(span)
                    r = span[0][0]
                    c = span[0][1]
                    if span_col_count > 1 and row == r and column == c:
                        text = table[r][c]
                        reflowed = reflow(text, 1)
                        min_lines = reflowed.split('\n')
                        min_text_length = len(max(min_lines, key=len))
                        add_col = 0
                        while sum(min_cell_widths[row][c:c + span_col_count]) < min_text_length:
                            min_cell_widths[row][c + add_col] += 1
                            if add_col + 1 < span_col_count:
                                add_col += 1
                            else:
                                add_col = 0
                        reflowed = reflow(text, 0)
                        max_lines = reflowed.split('\n')
                        max_text_length = len(max(max_lines, key=len))
                        add_col = 0
                        while sum(max_cell_widths[row][c:c + span_col_count]) < max_text_length:
                            max_cell_widths[row][c + add_col] += 1
                            if add_col + 1 < span_col_count:
                                add_col += 1
                            else:
                                add_col = 0
    
    max_col_widths = []
    min_col_widths = []

    for column in range(len(table[0])):
        max_col_widths.append(0)
        min_col_widths.append(0)
    for row in range(len(table)):
        for column in range(len(table[row])):
            if min_cell_widths[row][column] > min_col_widths[column]:
                min_col_widths[column] = min_cell_widths[row][column]
            if max_cell_widths[row][column] > max_col_widths[column]:
                max_col_widths[column] = max_cell_widths[row][column]

    column_count = len(table[0])
    text_space = space - (column_count + 1) - (2 * column_count)
    
    if sum(max_col_widths) <= text_space or space == 0:
        new_column_widths = max_col_widths
        
    elif sum(min_col_widths) > text_space:
        new_column_widths = min_col_widths

    else:
        max_col_widths = reevaluate_max_col_widths(
            max_col_widths, min_col_widths, space)
        cell_width_dicts, key_column_widths = get_key_column_widths(
            table, spans, max_col_widths, min_col_widths)
        
        col_height_width_dicts = []
        column = 0
        while column < column_count:
            long_row = has_column_span(table, column, spans)
            if not long_row:
                col_height_width_dicts.append(
                    make_column_height_width_dict(column, cell_width_dicts))
                column += 1
            else:
                height_width_dict, column = make_spanned_column_height_width_dict(
                    table, min_col_widths, column, long_row, spans, 
                    max_col_widths, cell_width_dicts, text_space)
                col_height_width_dicts.append(height_width_dict)
        
        col_heights = []
        for col in col_height_width_dicts:
            col_heights.append(list(col.keys()))
        
        combos = list(itertools.product(*col_heights))
        
        i = 0
        while i < len(combos):
            if sum(get_combo_widths(combos[i], col_height_width_dicts)) > text_space:
                combos.pop(i)
            else:
                i += 1
        
        smallest_height = grid_height_from_cols(table, spans, min_col_widths)
        min_widths = min_col_widths
        
        for i in range(len(combos)):
            height = grid_height_from_cols(table, spans, get_combo_widths(combos[i], col_height_width_dicts))
            if height < smallest_height:
                smallest_height = height
                min_widths = get_combo_widths(combos[i], col_height_width_dicts)
        new_column_widths = min_widths

    cell_widths = make_final_cell_widths(table, spans, new_column_widths)
    return cell_widths



from .tools import fromRoman

def is_enumerated_list_item(lines, index):
    '''Check if it's an enumerated list item'''
    if len(lines[index].lstrip()) == 0:
        return False
    
    enumerator = lines[index].lstrip().split(' ')[0]
    if not enumerator.endswith('.') and not enumerator.endswith(')'):
        return False
    
    enumerator = enumerator[0:-1]
    if enumerator.startswith('('):
        enumerator = enumerator[1::]
        
    try:
        int(enumerator)
        return True
    except:
        try:
            fromRoman(enumerator)
            return True
        except:
            if len(enumerator) == 1 and enumerator.isalpha():
                return True
            else:
                return False



def is_int(string):
    try:
        int(string)
        return True
    except ValueError:
        return False

def is_footnote(lines, index):
    if not lines[index].lstrip().startswith('.. ['):
        return False
    
    footnote_symbols = ['*', '†', '‡', '§', '¶', '#', '♠', '♥', '♦', '♣']
    if (not lines[index].lstrip()[4] in footnote_symbols and
        not is_int(lines[index].lstrip()[4])
    ):
        return False
    
    return True



import docutils.statemachine
import docutils.parsers.rst.tableparser
import itertools

def make_empty_table(row_count, column_count):
    table = []
    while row_count > 0:
        row = []
        for column in range(column_count):
            row.append('')
        table.append(row)
        row_count -= 1
    return table
    
def make_span(row, column, extra_rows, extra_columns):
    span = [[row, column]]
    for r in range(row, row + extra_rows + 1):
        span.append([r, column])
    for c in range(column, column + extra_columns + 1):
        span.append([row, c])
    span.append([r, c])
    
    return span

def grid2data(text):
    '''Convert Grid table to data (the kind used by Dashtable'''
    text = text.strip()
    lines = text.split('\n')
    for i in range(len(lines)):
        lines[i] = lines[i].strip()
        
    parser = docutils.parsers.rst.tableparser.GridTableParser()
    grid_data = parser.parse(docutils.statemachine.StringList(list(lines)))
    grid_data = list(grid_data)

    column_widths = grid_data.pop(0)
    column_count = len(column_widths)
    
    if len(grid_data[0]) > 0:
        use_headers = True
        headers = grid_data[0][0]
        row_count = len(grid_data[1]) + 1
        grid_data[1].insert(0, headers)
        grid_data.pop(0)
    else:
        use_headers = False
        grid_data.pop(0)
        row_count = len(grid_data[0])
        
    grid_data = grid_data[0]
    table = make_empty_table(row_count, column_count)
    spans = []
    
    for row in range(len(grid_data)):
        for column in range(len(grid_data[row])):
            try:
                text = '\n'.join(grid_data[row][column][3]).rstrip()
                table[row][column] = text
                extra_rows = grid_data[row][column][0]
                extra_columns = grid_data[row][column][1]
                span = make_span(row, column, extra_rows, extra_columns)
                span = sorted(span)
                span = list(span for span,_ in itertools.groupby(span))
                if not len(span) == 1:
                    spans.append(span)
            except TypeError:
                pass
    spans = sorted(spans)
    return table, spans, use_headers



from .tools import is_only

def is_gridtable(lines, i):
    if (
        lines[i].lstrip().startswith('+-') and 
        lines[i].rstrip().endswith('-+') and 
        is_only(lines[i].strip(), ['+', '-', ' '])
       ):
        return True
    return False



from .tools.space_fill import space_fill
from .reflow_paragraph import reflow_paragraph

def reflow_field(text, space):
    output = []
    
    leading_space = text.replace(text.lstrip(), '')
    
    field_name = ''
    words = text.strip().split(' ')
    for x in range(len(words)):
        if words[x].endswith(':') and not words[x].endswith('\\:'):
            field_name = ' '.join(words[0:x + 1])
            break
    
    rest_of_text = text[len(leading_space + field_name)::]

    interspace = rest_of_text.replace(rest_of_text.lstrip(), '')
    rest_of_text = rest_of_text.lstrip()
    
    lspace = leading_space + space_fill(len(field_name), ' ') + interspace
    
    paragraph = reflow_paragraph(rest_of_text, space, lspace)
    paragraph = paragraph.lstrip()
    
    lines = paragraph.split('\n')
    output.append(leading_space + field_name + interspace + lines[0])
    output.extend(lines[1::])

    return '\n'.join(output)



from .is_only import is_only
from .space_fill import space_fill
from .roman import fromRoman, toRoman
from .get_enumerator_level import get_enumerator_level
from .get_enumerator_type import get_enumerator_type
from .make_enumerator import make_enumerator
from .grid2data import grid2data
from .data2rst import data2rst



from .tools import is_only

def is_main_title(lines, index):
    '''
    Determine if the lines follow the format of a main title
    '''
    if index > len(lines) - 3:
        return False
    if len(lines[index]) == 0:
        return False

    symbol = lines[index][0]
    if is_only(lines[index], [symbol]):
        if is_only(lines[index + 2], [symbol]):
            if len(lines[index]) >= len(lines[index + 1]):
                if len(lines[index + 2]) >= len(lines[index + 1]):
                    return True
    return False



def is_option_list_item(lines, index):
    if not lines[index].startswith('-') and not lines[index].startswith('/'):
        return False
    
    leading_space = lines[index].replace(lines[index].lstrip(), '')
    
    if not '  ' in lines[index]:
        if index < len(lines) - 1: 
            if lines[index + 1].startswith(leading_space + ' '):
                return True
            else:
                return False
        else:
            return False
            
        
    return True



from setuptools import setup, find_packages

setup(
    name='reflowrst',
    packages=find_packages(),
    version='1.0.7',
    description='Modify valid rst text to fit within given width',
    long_description=open('README.rst').read(),
    author='doakey3',
    author_email='reflowrst.dmodo@spamgourmet.com',
    url='https://github.com/doakey3/reflowrst',
    download_url='https://github.com/doakey3/reflowrst/tarball/1.0.7',
    license='MIT',
)



def collect_directive(lines, index):
    output = []
    output.append(lines[index])
    index += 1
    while index < len(lines):
        if not lines[index] == '' and not lines[index].startswith(' '):
            return '\n'.join(output), index
        else:
            output.append(lines[index])
            index += 1
    return '\n'.join(output), index



def reflow_bullet_list_item(text, space):
    '''
    reflow a bullet list item
    '''
    if space == 0:
        return text
        
    leading_space = text.replace(text.lstrip(), '')
    line = text.lstrip()
    words = line.split(' ')
    growing_string = leading_space
    
    output_list = []
    while len(words) > 0:
        if len(growing_string) == len(leading_space):
            growing_string += words[0]
            words.pop(0)
        elif len(growing_string) == len(leading_space) + 1:
            growing_string += ' ' + words[0]
            words.pop(0)
        elif len(growing_string + ' ' + words[0]) <= space:
            growing_string += ' ' + words[0]
            words.pop(0)
        else:
            output_list.append(growing_string + '\n')
            growing_string = leading_space + ' '
    output_list.append(growing_string + '\n')
    return ''.join(output_list).rstrip()
